/***********************************************************************
 * Module : C3SPLINE.C - Provides routines to compute a curve fit on   *
 *                       the data for a profile graph                  *
 *                                                                     *
 * Programmer: Brian Davies - SAIC (original)                          *
 *             Kevin Russo - SAIC (1)                                  *
 *             Melissa Mathews - SAIC (2)                              *
 *                                                                     *
 *   Change     Date          Reason                                   *
 *   ------     ----          ------                                   *
 *   Original   ?             n/a                                      *
 *   1          07 Apr 93     modified for XVT and MUSE                *
 *   2          17 Jun 93     converted xvt 2.0 calls to xvt 3.1       *
 ***********************************************************************/

/****************
 * header files *
 ****************/

#include "los.h"

/**********************************
 * file scope function prototypes *
 **********************************/

#ifdef PROTO
static float f(float x);
#else
static float f();
#endif

/************************
 * file_scope variables *
 ************************/

static GHANDLE phdl, uhdl, dhdl, whdl;

static float *p, *u, *d, *w;

/************************************************************************
 *   FUNCTION     : init_cspline_handles
 *
 *   PURPOSE      : clears the cubic spline handles
 *
 *   INPUTS       : none
 *
 *   RETURN VALUE : void
 *
 *   CALLED FROM  : do_appl_init in module LOSAPPL.C
 *                  kill_cspline_handles in module C3SPLINE.C
 *
 ************************************************************************/           

#ifdef PROTO
void init_cspline_handles(void)
#else
void init_cspline_handles()
#endif
{
   phdl = (GHANDLE)NULL;
   uhdl = (GHANDLE)NULL;
   dhdl = (GHANDLE)NULL;
   whdl = (GHANDLE)NULL;
}

/************************************************************************
 *   FUNCTION     : kill_cspline_handles
 *
 *   PURPOSE      : release the memory pointed to by the cubic spline
 *                  handles
 *
 *   INPUTS       : none
 *
 *   RETURN VALUE : void
 *
 *   CALLED FROM  : do_file_menu in module LOSMENU1.C
 *
 ************************************************************************/           

#ifdef PROTO
void kill_cspline_handles(void)
#else
void kill_cspline_handles()
#endif
{
   if (phdl != (GHANDLE)NULL)
   {
      if (xvt_gmem_free(phdl) == FALSE)
         xvt_dm_post_fatal_exit("Unable to release memory in kill_cspline_handles()");
   }

   if (uhdl != (GHANDLE)NULL)
   {
      if (xvt_gmem_free(uhdl) == FALSE)
         xvt_dm_post_fatal_exit("Unable to release memory in kill_cspline_handles()");
   }

   if (dhdl != (GHANDLE)NULL)
   {
      if (xvt_gmem_free(dhdl) == FALSE)
         xvt_dm_post_fatal_exit("Unable to release memory in kill_cspline_handles()");
   }

   if (whdl != (GHANDLE)NULL)
   {
      if (xvt_gmem_free(whdl) == FALSE)
         xvt_dm_post_fatal_exit("Unable to release memory in kill_cspline_handles()");
   }

   init_cspline_handles();
}

/************************************************************************
 *   FUNCTION     : makespline
 *
 *   PURPOSE      : builds a cubic spline for a given set of data
 *
 *   INPUTS       : scrn_pts - a set of screen coords to perform spline on
 *                  num_pts  - number of screen coords
 *
 *   RETURN VALUE : none
 *
 *   CALLED FROM  : construct_profile() in module BUILD_PR.C
 *
 *   CALLS        : none
 *
 ************************************************************************/

#ifdef PROTO
void makespline (PNT *scrn_pts, long num_pts)
#else
void makespline (scrn_pts, num_pts)
PNT  *scrn_pts;
long num_pts;
#endif
{
   long i;

   if ((phdl != (GHANDLE)NULL) && (uhdl != (GHANDLE)NULL))
      return;

   phdl = (GHANDLE)xvt_gmem_alloc((num_pts+1)*sizeof(float));
   uhdl = (GHANDLE)xvt_gmem_alloc((num_pts+1)*sizeof(float));
   dhdl = (GHANDLE)xvt_gmem_alloc((num_pts+1)*sizeof(float));
   whdl = (GHANDLE)xvt_gmem_alloc((num_pts+1)*sizeof(float));

   if ((phdl == (GHANDLE)NULL) || (uhdl == (GHANDLE)NULL) ||
       (dhdl == (GHANDLE)NULL) || (whdl == (GHANDLE)NULL))
      xvt_dm_post_fatal_exit("unable to allocate memory in makespline()");

   p = (float *)xvt_gmem_lock(phdl);
   u = (float *)xvt_gmem_lock(uhdl);
   d = (float *)xvt_gmem_lock(dhdl);
   w = (float *)xvt_gmem_lock(whdl);

   if ((p == (float *)NULL) || (u == (float *)NULL) ||
       (d == (float *)NULL) || (w == (float *)NULL))
      xvt_dm_post_fatal_exit("unable to lock down memory in makespline()");

   for (i = 2; i < num_pts ; i++)
      d[i] = 2 * (float)(scrn_pts[i+1].h - scrn_pts[i-1].h);

   for (i = 1; i < num_pts; i++)
      u[i] = (float)scrn_pts[i+1].h - (float)scrn_pts[i].h;

   for (i = 2; i < num_pts; i++)
      w[i] = 6 * (float)(scrn_pts[i+1].v - scrn_pts[i].v) / u[i] -
                 (float)(scrn_pts[i].v - scrn_pts[i-1].v) / u[i-1];

   p[1]       = (float)ZERO;
   p[num_pts] = (float)ZERO;

   for (i = 2; i < num_pts-1; i++)
   {
      w[i+1] = w[i+1] - w[i] * u[i] / d[i];
      d[i+1] = d[i+1] - u[i] * u[i] / d[i];
   }

   for(i = num_pts-1; i > 1; i--)
      p[i] = (w[i] - u[i] * p[i+1]) / d[i];

   if (xvt_gmem_unlock(phdl) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock memory in makespline()");
   if (xvt_gmem_unlock(uhdl) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock memory in makespline()");
   if (xvt_gmem_unlock(dhdl) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock memory in makespline()");
   if (xvt_gmem_unlock(whdl) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock memory in makespline()");

   if (xvt_gmem_free(dhdl) == FALSE)
      xvt_dm_post_fatal_exit("Unable to release memory in makespline()");
   if (xvt_gmem_free(whdl) == FALSE)
      xvt_dm_post_fatal_exit("Unablt to release memory in makespline()");

   dhdl = (GHANDLE)NULL;
   whdl = (GHANDLE)NULL;
}

/************************************************************************
 *   FUNCTION     : eval
 *
 *   PURPOSE      : determines a data point (y) based on some (x) in the
 *                  cubic spline, essentially performs y = f(x)
 *
 *   INPUTS       : scrn_pts - a set of screen coords to use in calcs
 *                  num_pts  - number of screen coords
 *                  x        - the data point to calculate y from
 *
 *   RETURN VALUE : the answer to the function y = f(x)
 *
 *   CALLED FROM  : construct_profile in module BUILD_PR.C
 *
 *   CALLS        : f() in module C3SPLINE.C
 *
 ************************************************************************/

#ifdef PROTO
short eval(PNT *scrn_pts, short x)
#else
short eval(scrn_pts, x)
PNT    *scrn_pts;
short x;
#endif
{
   float t, y;
   long   i = 1;

   p = (float *)xvt_gmem_lock(phdl);
   u = (float *)xvt_gmem_lock(uhdl);

   if ((p == (float *)NULL) || (u == (float *)NULL))
      xvt_dm_post_fatal_exit("unable to lock down memory in eval()");

   while (x > scrn_pts[i+1].h)
      i++;

   t = (float)(x - scrn_pts[i].h) / u[i];

   y = t * (float)scrn_pts[i+1].v + (1-t) *
       (float)scrn_pts[i].v + u[i] * u[i] *
       (f(t) * p[i+1] + f(1-t) * p[i]) / 6;

   if (xvt_gmem_unlock(phdl) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock memory in eval()");
   if (xvt_gmem_unlock(uhdl) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock memory in eval()");

   return (short)(y + 0.5);
}

/************************************************************************
 *   FUNCTION     : f
 *
 *   PURPOSE      : returns a number cubed minus itself
 *
 *   INPUTS       : x - the data point used in calcs 
 *
 *   RETURN VALUE : x cubed minus x
 *
 *   CALLED FROM  : eval() in module C3SPLINE.C
 *
 ************************************************************************/

#ifdef PROTO
static float f(float x)
#else
static float f(x)
float x;
#endif
{
   return (x*x*x - x);
}
